import pandas as pd
import numpy as np
from binance.client import Client
from binance.enums import HistoricalKlinesType
from ta.volatility import average_true_range


def download_data(ticket: str, ts_start: int, ts_end: int = 0,
                  timeframe: str = Client.KLINE_INTERVAL_1MINUTE) -> pd.DataFrame:
    """ Скачать данные с Binance API """
    client = Client('', '')
    titles = ['open_time', 'open', 'high', 'low', 'close', 'volume', 'close_time', 'Quote asset volume',
              'Number of trades', 'Taker buy base asset volume', 'Taker buy quote asset volume', 'Can be ignored']
    data_generator = client.get_historical_klines_generator(
        symbol=ticket,
        interval=timeframe,
        start_str=ts_start * 1000,
        end_str=None if ts_end == 0 else ts_end * 1000,
        klines_type=HistoricalKlinesType.FUTURES
    )
    rows = []
    for row in data_generator:
        rows.append(row)
    df = pd.DataFrame(rows, columns=titles)
    df = df[['close_time', 'close', 'high', 'low', 'volume']]
    df[['close', 'high', 'low', 'volume']] = df[['close', 'high', 'low', 'volume']].astype(np.float64)
    return df


def calc_rel_atr(df: pd.DataFrame, atr_window: int) -> pd.Series:
    """ Расчитать относительный АТР """
    return average_true_range(df['high'], df['low'], df['close'], window=atr_window) / df['close'] * 100


def cal_rel_atr_by_other_timeframe(df: pd.DataFrame, ticket: str, atr_window: int,
                                   timeframe: str = Client.KLINE_INTERVAL_1DAY) -> pd.Series:
    """ Получить дневной атр """
    period_seconds = {
        Client.KLINE_INTERVAL_1DAY: 24 * 60 * 60,
        Client.KLINE_INTERVAL_4HOUR: 4 * 60 * 60,
        Client.KLINE_INTERVAL_1HOUR: 1 * 60 * 60,
        Client.KLINE_INTERVAL_30MINUTE: 30 * 60,
        Client.KLINE_INTERVAL_15MINUTE: 15 * 60
    }[timeframe]
    df_start_date = int(df['close_time'][0] / 1000)
    df_end_date = int(df['close_time'][len(df) - 1] / 1000)
    data = download_data(ticket, df_start_date - (atr_window + 1) * period_seconds, df_end_date, timeframe)
    data['atr'] = calc_rel_atr(data, atr_window)
    atr_by_days = {int(row['close_time']): row['atr'] for i, row in data.iterrows()}
    del data
    atr_arr = []
    for i, row in df.iterrows():
        for ts, atr in atr_by_days.copy().items():
            diff = (row['close_time'] - ts) / 1000
            if diff > period_seconds:
                del atr_by_days[ts]
            else:
                atr_arr.append(atr)
                break
    return pd.Series(atr_arr)


import itertools as it
from time import sleep
from typing import Callable, Dict, Tuple, Type
from math import inf
from multiprocess import Pool
import numpy as np
import pandas as pd
import nevergrad as ng
import matplotlib.pyplot as plt
from .Strategy import Strategy


def calc_efficiency(train_profit: float, test_profit: float) -> float:
    if train_profit <= 0:
        return np.nan
    return round((test_profit / train_profit) * 100, 2)


def get_calc_strategy_func(strategy: Type[Strategy], brut=False, direction=Strategy.LONG) -> Callable:
    """ Получить функцию для расчета минимумов по стратегии """
    s = strategy

    if brut:
        def calc_strategy_profit(df, params) -> float:
            s_obj = s()
            s_obj.tune(for_opt=True, direction=direction)
            s_obj.run(df, params)
            return -s_obj.coef
    else:
        def calc_strategy_profit(df, **params) -> float:
            s_obj = s()
            s_obj.tune(for_opt=True, direction=direction)
            s_obj.run(df, params)
            return -s_obj.coef
    return calc_strategy_profit


def find_best_params_optimization(
        calc_strategy: Callable,
        df: pd.DataFrame,
        param_ranges: Dict[str, list]
) -> Tuple[float, Dict[str, float]]:
    """ Поиск лучшех параметров подбором с оптимизацией"""
    params = {}
    combinations = 0
    for param_name in param_ranges:
        if combinations == 0:
            combinations = len(param_ranges[param_name])
        else:
            combinations *= len(param_ranges[param_name])
        params[param_name] = ng.p.TransitionChoice(param_ranges[param_name])
    parametrization = ng.p.Instrumentation(
        df=ng.p.TransitionChoice([df]),
        **params
    )
    budget = int(combinations * 0.25)
    budget = max(80, budget)
    optimizer = ng.optimizers.NGOpt(parametrization=parametrization, budget=budget)
    recommendation = optimizer.minimize(calc_strategy)
    best_params = recommendation.kwargs
    del best_params['df']
    best_coef = calc_strategy(df, **best_params)
    return -best_coef, best_params


def find_best_params_brut_parallel(
        calc_strategy: Callable,
        df: pd.DataFrame,
        param_ranges: Dict[str, list]
) -> Tuple[float, Dict[str, float]]:
    """ Поиск лучшех параметров подбором с параллелизацией """
    param_ranges_list = []
    for param_range in param_ranges.values():
        param_ranges_list.append(param_range)
    combs = list(it.product(*param_ranges_list))
    param_names = list(param_ranges.keys())
    best_coef = inf
    best_params = {}
    tasks = {}
    params = {}
    with Pool(12) as pool:
        for i, comb in enumerate(combs):
            args = dict(zip(param_names, comb))
            tasks[i] = pool.apply_async(calc_strategy, (df, args))
            params[i] = args
        while len(tasks):
            for i, task in tasks.copy().items():
                if task.ready():
                    coef = task.get()
                    del tasks[i]
                    if coef < best_coef:
                        best_coef = coef
                        best_params = params[i]
    return -best_coef, best_params


def find_best_params_brut(
        calc_strategy: Callable,
        df: pd.DataFrame,
        param_ranges: Dict[str, list]
) -> Tuple[float, Dict[str, float]]:
    """ Поиск лучшех параметров подбором """
    param_ranges_list = []
    for param_range in param_ranges.values():
        param_ranges_list.append(param_range)
    combs = list(it.product(*param_ranges_list))
    param_names = list(param_ranges.keys())
    best_coef = inf
    best_params = {}
    for comb in combs:
        args = dict(zip(param_names, comb))
        coef = calc_strategy(df, args)
        if coef < best_coef:
            best_coef = coef
            best_params = args
    return -best_coef, best_params


class Epoch:
    """ Данные по эпохе """

    def __init__(self):
        self.task = None

        self.df_train: pd.DataFrame = None
        self.df_test: pd.DataFrame = None

        self.strategy_train: Strategy = None
        self.strategy_test: Strategy = None

        self.best_params = {}
        self.is_done = False

    def get_train_ts(self) -> Dict[str, int]:
        """ Получить временные метки эпохи обучения """
        return {
            'start': int(self.df_train['close_time'][0] / 1000),
            'end': int(self.df_train['close_time'][len(self.df_train) - 1] / 1000)
        }

    def get_test_ts(self) -> Dict[str, int]:
        """ Получить временные метки эпохи теста """
        return {
            'start': int(self.df_test['close_time'][0] / 1000),
            'end': int(self.df_test['close_time'][len(self.df_test) - 1] / 1000)
        }


class FakeTask:
    """ Фейковая задача """

    def __init__(self, func, params):
        self.func = func
        self.params = params

    def ready(self):
        return True

    def get(self):
        return self.func(*self.params)


class BaseForward:
    """ Форвардный анализ """

    def __init__(self):
        self.result: pd.DataFrame = None

        self.df: pd.DataFrame = None
        self.param_ranges = {}
        self.strategy: Type[Strategy] = None

        self.epoch_strategies = {}

        self.train_window = 0
        self.test_window = 0
        self.train_window_days = 0
        self.test_window_days = 0

        self.epochs_data: Dict[int, Epoch] = {}

        self.pool: Pool = None
        self.done_tasks = 0

        self.direction = Strategy.LONG

    def run(self, strategy: Type[Strategy], df: pd.DataFrame, param_ranges: Dict[str, list],
            train_window_days: int, test_window_days: int):
        """ Запустить форвардный анализ """
        self.df = df
        self.param_ranges = param_ranges
        self.strategy = strategy

        self.result = pd.DataFrame(
            columns=['epoch', 'coef_eff', 'train_rel_profit', 'test_profit'] + list(param_ranges.keys())
        )

        self.train_window_days = train_window_days
        self.test_window_days = test_window_days

        self.train_window = train_window_days * 24 * 60
        self.test_window = test_window_days * 24 * 60

        self.init_epochs_data()
        self.done_tasks = 0

        self.pool = Pool(processes=12)
        try:
            self.create_tasks()
            self.monitor_tasks()
        finally:
            self.pool.close()

    def init_epochs_data(self):
        """ Инициализировать данные по эпохам """
        epochs = int((len(self.df) - self.train_window) / self.test_window)
        for epoch in range(epochs):
            self.epochs_data[epoch] = Epoch()

            epoch_start_learn = epoch * self.test_window
            epoch_end_learn = epoch_start_learn + self.train_window
            self.epochs_data[epoch].df_train = self.df[epoch_start_learn:epoch_end_learn].reset_index(drop=True)

            epoch_start_test = epoch_end_learn
            epoch_end_test = epoch_start_test + self.test_window
            self.epochs_data[epoch].df_test = self.df[epoch_start_test:epoch_end_test].reset_index(drop=True)

    def create_tasks(self):
        pass

    def monitor_tasks(self):
        """ Мониторить задачи """
        total_tasks = len(self.epochs_data)
        while self.done_tasks < total_tasks:
            print("\r ", end="")
            print(f' leave epochs: {self.done_tasks}/{total_tasks}'.ljust(40), end="")
            for epoch, epoch_data in self.epochs_data.items():
                if epoch_data.is_done is False and epoch_data.task.ready():
                    self.done_tasks += 1
                    self.epochs_data[epoch].is_done = True

                    train_coef, best_params = epoch_data.task.get()
                    self.epochs_data[epoch].best_params = best_params

                    test_strategy = self.strategy()
                    test_strategy.run(epoch_data.df_test, best_params)
                    test_profit = test_strategy.get_profit()

                    train_strategy = self.strategy()
                    train_strategy.run(epoch_data.df_train, best_params)
                    train_profit = train_strategy.get_profit()
                    train_rel_profit = train_profit * self.test_window_days / self.train_window_days

                    self.result = self.result.append({
                        'epoch': epoch,
                        'train_coef': train_coef,
                        'train_rel_profit': train_rel_profit,
                        'test_profit': test_profit,
                        'coef_eff': calc_efficiency(train_rel_profit, test_profit),
                        **best_params
                    }, ignore_index=True)
                    self.epochs_data[epoch].strategy_test = test_strategy
                    self.epochs_data[epoch].strategy_train = train_strategy
                    print("\r ", end="")
                    print(f' leave epochs: {self.done_tasks}/{total_tasks}'.ljust(40), end="")
            sleep(0.1)

        print("\r ", end="")
        print(f' leave epochs: {self.done_tasks}/{total_tasks}'.ljust(40))
        self.result = self.result.sort_values('epoch').reset_index(drop=True)

    def get_epoch(self, number: int) -> Epoch:
        """ Получить эпоху """
        return self.epochs_data[number]

    def draw_train_profit(self):
        plt.plot(self.result['train_rel_profit'].cumsum())
        plt.show()

    def draw_test_profit(self):
        plt.plot(self.result['test_profit'].cumsum())
        plt.show()

    def draw_profit(self):
        plt.plot(self.result['test_profit'].cumsum())
        plt.plot(self.result['train_rel_profit'].cumsum())
        plt.show()

    def calc_full_test(self) -> pd.DataFrame:
        trades = pd.DataFrame(
            columns=['open_time', 'close_time', 'av_price', 'close',
                     'volume', 'profit', 'max_drawdown', 'max_possible_profit']
        )
        last_stop = 0
        for epoch in self.epochs_data.values():
            start, stop = epoch.get_test_ts().values()
            start *= 1000
            stop *= 1000
            start = max(start, last_stop)
            strategy = self.strategy()
            strategy.run(self.df, epoch.best_params)
            epoch_trades = strategy.trades
            epoch_trades = epoch_trades[(epoch_trades['open_time'] >= start) & (epoch_trades['open_time'] <= stop)]
            trades = trades.append(epoch_trades, ignore_index=True)
            total_trades = len(trades)
            if total_trades:
                last_stop = trades['open_time'][total_trades - 1] + 1
        return trades


class ForwardOpt(BaseForward):
    """ Форвардный анализ с оптимизацией, паралельный просчет эпох """

    def create_tasks(self):
        """ Создать задачи """
        for epoch, epoch_data in self.epochs_data.copy().items():
            self.epochs_data[epoch].task = self.pool.apply_async(find_best_params_optimization, (
                get_calc_strategy_func(self.strategy, direction=self.direction),
                epoch_data.df_train,
                self.param_ranges
            ))


class ForwardBrut(BaseForward):
    """ Форвардный анализ, паралельный просчет эпох """

    def create_tasks(self):
        """ Создать задачи """
        for epoch, epoch_data in self.epochs_data.copy().items():
            self.epochs_data[epoch].task = self.pool.apply_async(find_best_params_brut, (
                get_calc_strategy_func(self.strategy, brut=True, direction=self.direction),
                epoch_data.df_train,
                self.param_ranges
            ))


class ForwardParallelBrut(BaseForward):
    """ Форвардный анализ, паралельный подбор параметров """

    def create_tasks(self):
        """ Создать задачи """
        for epoch, epoch_data in self.epochs_data.copy().items():
            self.epochs_data[epoch].task = FakeTask(find_best_params_brut_parallel, [
                get_calc_strategy_func(self.strategy, brut=True, direction=self.direction),
                epoch_data.df_train,
                self.param_ranges
            ])




import math
from functools import reduce
import pandas as pd
import matplotlib.pyplot as plt


class Strategy():
    """ Стратегия """

    LONG = 'long'
    SHORT = 'short'

    def __init__(self):

        # Датафрейм для обсчета
        self.df: pd.DataFrame = None

        # Датафрейм со сделками
        self.trades: pd.DataFrame = None

        # Временный массив под сделки
        self.trades_temp = []

        # Временный мессив с позициями
        self.position_value_arr = []

        # Данные по ценам и времени открытия по сделкам текущей позиции
        self.current_position_opens = []

        # Количество сделок в текущей позиции
        self.position = 0

        # Средняя цена позиции
        self.av_price = 0

        # Просадка по позиции
        self.drawdown = math.inf

        # Максимальный профит по позиции
        self.possible_profit = -math.inf

        # Коэф анализу
        self.coef = 0

        # Объем лота сделки в доллагах
        self.lot_dollars = 10

        # Направление торговли
        self.direction = self.LONG

        # Запуск только для оптимизации
        self.for_opt = False

    def tune(self, lot_dollars: float = None, direction: str = None, for_opt: bool = None):
        """ Установить настройки """
        for setting_name, value in locals().items():
            if value is not None:
                self.__setattr__(setting_name, value)

    def increase_position(self, row):
        """ Увеличить позицию """
        self.position += 1
        self.current_position_opens.append({
            'time': row.close_time,
            'price': row.close
        })
        self.av_price = reduce(lambda s, x: s + x['price'], self.current_position_opens, 0) / self.position

    def close_position(self, row):
        """ Закрыть позицию """
        profit = self.calc_trade_profit(self.av_price, row.close, self.position)
        self.trades_temp.append({
            'open_time': self.current_position_opens[0]['time'],
            'close_time': row.close_time,
            'av_price': self.av_price,
            'close': row.close,
            'volume': self.position,
            'profit': profit,
            'max_drawdown': self.drawdown,
            'max_possible_profit': self.possible_profit
        })
        self.coef += self.calc_trade_coef(profit)
        self.reset_position_state()

    def calc_trade_profit(self, open_price: float, close_price: float, volume: float) -> float:
        """ Сосчитать профит по сделке """
        diff = close_price - open_price
        if self.direction == self.SHORT:
            diff = -diff
        return self.lot_dollars * volume / open_price * (diff - 0.0004 * (close_price + open_price))

    def reset_position_state(self):
        """ Сбросить состояние между позициями """
        self.current_position_opens = []
        self.position = 0

        self.drawdown = math.inf
        self.possible_profit = -math.inf

    def reset_process_state(self):
        """ Сбросить состояние между обсчетами"""
        self.trades = None

        self.trades_temp = []
        self.position_value_arr = []

        self.coef = 0

    def run(self, df: pd.DataFrame, params: dict):
        """ Запустить расчеты """
        self.reset_process_state()
        self.reset_position_state()
        self.process(df, params)
        self.set_data_after_process(df)

    def set_data_after_iter(self, row):
        """ Рассчитать максимальные диапазоны профита внутри сделки """
        if self.position > 0:
            current_profit = self.calc_trade_profit(self.av_price, row.close, self.position)
            self.drawdown = min(current_profit, self.drawdown)
            self.possible_profit = max(current_profit, self.drawdown)
        self.position_value_arr.append(self.position)

    def set_data_after_process(self, df: pd.DataFrame):
        if self.for_opt is False:
            self.df = df.copy()
            self.df['position'] = self.position_value_arr
            self.trades = pd.DataFrame(
                self.trades_temp,
                columns=['open_time', 'close_time', 'av_price', 'close',
                         'volume', 'profit', 'max_drawdown', 'max_possible_profit']
            )

    def process(self, df: pd.DataFrame, params: dict):
        """ Обработать датафрем по тратегии """
        pass

    def get_profit(self) -> float:
        """ Получить профит по стратегии """
        return self.trades['profit'].sum()

    @staticmethod
    def calc_trade_coef(profit) -> float:
        """ Рассчитать COEF по сделке """
        if profit < 0:
            return profit + 1.05 ** profit - 1
        return profit

    def draw_cum(self):
        """ Отрисовать кумуляту сделок """
        plt.plot(self.trades['close_time'].values, self.trades['profit'].cumsum().values)
        plt.show()

    def draw_cum_and_price(self):
        """ Отрисовать кумуляту сделок и цену"""
        fig, ax1 = plt.subplots()
        ax2 = ax1.twinx()
        ax1.plot(self.df['close_time'].values, self.df['close'].values, label='Цена', color='darkred')
        ax2.plot(self.trades['close_time'].values, self.trades['profit'].cumsum().values, label='Прибыль')
        ax1.set_ylabel('Цена')
        ax2.set_ylabel('Прибыль')
        fig.legend()
        plt.show()

    @staticmethod
    def draw_positions(df, need_av_enter=True):
        """ Отрисовать позиции на датафрейме """
        df = df.reset_index(drop=True)
        df['pos_diff'] = df['position'].diff()
        trade_inputs = df[df['pos_diff'] > 0]
        trade_outputs = df[df['pos_diff'] < 0]
        plt.plot(df['close_time'].values, df['close'].values, color='orange')
        plt.plot(df['close_time'].values, df['ma'].values)
        if need_av_enter:
            av_price = trade_inputs['close'].mean()
            plt.plot(
                [df['close_time'][0], df['close_time'][len(df) - 1]],
                [av_price, av_price],
                linestyle='--',
                color='#9EF5A1'
            )
        plt.scatter(trade_inputs['close_time'].values, trade_inputs['close'].values, marker="^", c='green', zorder=10)
        plt.scatter(trade_outputs['close_time'].values, trade_outputs['close'].values, marker="v", c='red', zorder=10)
        plt.show()

    def get_df_by_trade_index(self, trade_index: int) -> pd.DataFrame:
        """ Получить датафрейм по индексу сделки """
        start_ts = self.trades['open_time'][trade_index]
        end_ts = self.trades['close_time'][trade_index]
        return self.df[(self.df['close_time'] >= start_ts) & (self.df['close_time'] <= end_ts)]



from setuptools import setup

setup(
    version='0.2.5',
    name = 'tradeanal',
    description='tradeanal',
    author='cherepanov_max95',
    author_email='cherepanov@ku66.ru',
    install_requires=[
        'multiprocess',
        'nevergrad',
        'pandas',
        'matplotlib'
    ],
    include_package_data=True,
    zip_safe=False,
    packages=['tradeanal']
)